#include <stdio.h>
#include <stdarg.h>
#include <cmath>

#include "dprintf.hh"

static const unsigned colors[] =
{
    0115,
    0305,
    0010,
    0151,
    0522,
    0330
};

static double RgbDiff(unsigned rgb1, unsigned rgb2)
{
    unsigned R1 = ((rgb1>>16)&0xFF), R2 = ((rgb2>>16)&0xFF);
    unsigned G1 = ((rgb1>> 8)&0xFF), G2 = ((rgb2>> 8)&0xFF);
    unsigned B1 = ((rgb1>> 0)&0xFF), B2 = ((rgb2>> 0)&0xFF);
    int Y1=R1,U1=G1,V1=B1, Y2=R2,U2=G2,V2=B2;
    /*
    double Y1 = (( ( 66 * R1 + 129 * G1 + 25 * B1 + 128) / 256.0) + 16);
    double U1 = ( ( -38 * R1 - 74 * G1 + 112 * B1 + 128) / 256.0) + 128;
    double V1 = ( ( 112 * R1 - 94 * G1 - 18 * B1 + 128) / 256.0) + 128;
    double Y2 = (( ( 66 * R2 + 129 * G2 + 25 * B2 + 128) / 256.0) + 16);
    double U2 = ( ( -38 * R2 - 74 * G2 + 112 * B2 + 128) / 256.0) + 128;
    double V2 = ( ( 112 * R2 - 94 * G2 - 18 * B2 + 128) / 256.0) + 128;
    */
    return std::pow((Y1-Y2),2.0)
         + std::pow((U1-U2),2.0)
         + std::pow((V1-V2),2.0);
}

static unsigned CreateXtermColor(unsigned rgb)
{
    #define sixtable(n)  (n ? 55+(n)*40 : 0)
    #define graytable(n) (8 + 10*(n))  /* n=0..23 */
    unsigned r = (rgb>>16)&0xFF, g = (rgb>>8)&0xFF, b = (rgb>>0)&0xFF;

    // Try six-algorithm
    unsigned r6 = r*6/256, g6=g*6/256, b6=b*6/256;
    unsigned six_result = (sixtable(r6)<<16) + (sixtable(g6)<<8) + (sixtable(b6));

    // Try grayscale-algorithm
    unsigned gray_value  = (r*30+g*60+b*11)/101; // approximate Y formula
    gray_value = (gray_value<8) ? 0 : (gray_value>0xEE ? 23 : (gray_value-8+5)/10);
    unsigned gray_result = graytable(gray_value);

    if(RgbDiff(gray_result,rgb) < RgbDiff(six_result,rgb))
    {
        // Select grayscale value
        return 232 + gray_value;
    }

    // Of the first 16 ansi colours, only black we can trust
    // that it's some certain value. The others vary wildly.
    if(six_result == 0) return 0; // black

    // Select six-cube value.
    return 16 + b6 + g6*6 + r6*6*6;
}

void dprintf(dprintf_type type,const char*fmt,...)
{
    unsigned color = colors[type];

    unsigned xc = CreateXtermColor(
        (((color&0700)*255/0500)<<16) +
        (((color&0070)*255/0050)<<8) +
        (((color&0007)*255/0005)));
    if(xc < 8)
        fprintf(stderr, "\33[%um", 30+xc);
    else
        fprintf(stderr, "\33[38;5;%um", xc);

    va_list ap;
    va_start(ap,fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, "\33[0m");
}
