#include <cctype>

#include "data.hh"

template<typename T, bool FloatType, bool UnsignedType, unsigned long long minvalue, unsigned long long maxvalue>
void B_ReadGeneric(T& target, const char* input)
{
    long double         longdouble_result;
    unsigned long long  longlong_result;
    bool negative           = false;
    bool definitely_integer = false;
    unsigned integer_base   = 10;

    while(std::isspace(*input)) ++input;

    if(*input == '-') { negative = true; ++input; }
    if(*input == '+') ++input;

    if(*input == '&' && (input[1] == 'H' || input[1] == 'h'))
        { integer_base = 16; definitely_integer = true; input += 2; }
    else if(*input == '&' && (input[1] == 'O' || input[1] == 'o'))
        { integer_base =  8; definitely_integer = true; input += 2; }
    else if(*input == '&' && (input[1] == 'B' || input[1] == 'b'))
        { integer_base =  2; definitely_integer = true; input += 2; }
    else if(*input == '0' && (input[1] == 'X' || input[1] == 'x'))
        { integer_base = 16; definitely_integer = true; input += 2; }
    else if(*input == '0' && (input[1] == 'B' || input[1] == 'b'))
        { integer_base =  2; definitely_integer = true; input += 2; }

    if(!*input) { err: B_ERROR(2); return; } // SYNTAX ERROR

    char*end = 0;
    longlong_result = std::strtol(input, &end, integer_base);

    if(!definitely_integer && (*input == '.' || *input == 'e' || *input == 'E'))
    {
        longdouble_result = std::strtold(input, &end);
        definitely_integer = false;
    }
    else
        definitely_integer = true;

    while(std::isspace(*end)) ++end;
    if(*end) goto err;

    switch(negative*2 + definitely_integer)
    {
        case 0: // not negative, not integer
            target = longdouble_result;
            if(!FloatType && longdouble_result > maxvalue) B_ERROR(6); // overflow
            break;
        case 1: // not negative, integer
            target = longlong_result;
            if(!FloatType && longlong_result > maxvalue) B_ERROR(6); // overflow
            break;
        case 2: // negative, not integer
            target = -longdouble_result;
            if(!FloatType && longdouble_result > -minvalue) B_ERROR(6); // overflow
            break;
        case 3: // negative, integer
            target = -longlong_result;
            if(!FloatType && longlong_result > -minvalue) B_ERROR(6); // overflow
            break;
    }
}

void B_READ_DO(BAS_DBL& target, const char* input)
{
    B_ReadGeneric<BAS_DBL, true, false, 0,0> (target, input);
}

void B_READ_DO(BAS_SGN& target, const char* input)
{
    B_ReadGeneric<BAS_SGN, true, false, 0,0> (target, input);
}

void B_READ_DO(BAS_INT& target, const char* input)
{
    B_ReadGeneric<BAS_INT, false, false, 0x80000000u, 0x7FFFFFFF  > (target, input);
}

void B_READ_DO(BAS_LNG& target, const char* input)
{
    B_ReadGeneric<BAS_LNG, false, false, 0x80000000u, 0x7FFFFFFF > (target, input);
}

void B_READ_DO(BAS_BYTE& target, const char* input)
{
    B_ReadGeneric<BAS_BYTE, false, false, 0x80, 0x7F > (target, input);
}

void B_READ_DO(BAS_UINT& target, const char* input)
{
    B_ReadGeneric<BAS_UINT, false, true, 0, BAS_UINT(~BAS_UINT(0))> (target, input);
}

void B_READ_DO(BAS_ULONG& target, const char* input)
{
    B_ReadGeneric<BAS_ULONG, false, true, 0, BAS_ULONG(~BAS_ULONG(0)) > (target, input);
}

void B_READ_DO(BAS_UBYTE& target, const char* input)
{
    B_ReadGeneric<BAS_UBYTE, false, true, 0, BAS_UBYTE(~BAS_UBYTE(0)) > (target, input);
}
