#include "input.hh"
#include "bstringfun.hh"
#include "gfx.hh"

void InputImplement::InputTo(BAS_INT& target)
{
    long l = InputLong(); target = l;
    if(l < -32768 || l > 32767) GotOverflow();
}

void InputImplement::InputTo(BAS_SGN& target)
{
    target = InputLong();
}

void InputImplement::InputTo(BAS_DBL& target)
{
    target = InputLong();
}

void InputImplement::InputTo(BAS_LNG& target)
{
    long double l = InputLong(); target = l;
    if(l < -2147483648.0
    || l > 2147483647.0) GotOverflow();
}

void InputImplement::InputTo(BAS_STR& target)
{
    // anything until comma or eol or eof
    // ignore spaces at beginning and end
    // if quote encountered while string empty, parse until quote or eol or eof.
    //    After quote, ignore spaces.
    target = "";
    bool quotemode = false;
    bool hadquote  = false;
    for(;;)
    {
        int ch = InputChar();
        if(ch == -1) break;
        if((target.empty() || hadquote) && (ch == ' ' || ch == '\t')) continue;
        
        /*fprintf(stderr, "gotc(%c)\n", ch);*/
        
        if(ch == '"' && target.empty()) { quotemode = true; continue; }
        if(ch == '"' && quotemode) { quotemode=false; hadquote=true; continue; }
        if(ch == ',' && !quotemode) break;
        if(hadquote) { InputError = true; break; }
        target += (char)ch;
    }
    target = RTrim(target);

    /*fprintf(stderr, "target(%s)err(%d)\n",
        target.c_str(), (int)InputError);*/
}

long double InputImplement::InputLong()
{
    // anything until comma or eol or eof
    // ignore spaces at beginning and end
    
    std::string target;
    for(;;)
    {
        int ch = InputChar();
        if(ch == -1) break;
        if(target.empty() && (ch == ' ' || ch == '\t')) continue;
        
        if(ch == ',') break;
        target += (char)ch;
    }
    
    char* endptr;
    const char* beginptr = target.c_str();
    long double res = strtold(beginptr, &endptr);
    /*fprintf(stderr, "res(%lf)begin(%s)end(%s)\n",
        (double)res,
        beginptr,
        endptr);*/
    
    if(endptr == beginptr) { InputError = true; return 0.0; }
    while(*endptr)
    {
        if(*endptr != ' ' && *endptr != '\t') { InputError = true; break; }
        ++endptr;
    }
    return res;
}

bool InputImplement::FinishInput()
{
    while(!GotEOL)
    {
        int ch = InputChar();
        if(ch == -1) break;
        if(ch != ' ' || ch != '\t')
        {
            //fprintf(stderr, "Finish: '%c'?\n", ch);
            InputError = true; return false;
        }
    }
    return InputError;
}
