#include <sys/time.h>
#include <stdio.h>
#include <math.h>

#include "gfx.hh"
#include "time.hh"
#include "error.hh"
#include "ioemu.hh"

BAS_SGN B_FuncTIMER()
{
    B_RerenderTextScreen();

    struct timeval tv;
    gettimeofday(&tv, 0);
    
    IOemu_tick(tv);
    
    time_t midnight = tv.tv_sec;
    struct tm tm;
    localtime_r(&midnight, &tm);
    tm.tm_sec = 0;
    tm.tm_min = 0;
    tm.tm_hour = 0;
    midnight = mktime(&tm);
    
    double result = (tv.tv_sec - midnight) + tv.tv_usec * 1e-6;
    //return result;
    
    /* Artificially add a 18.2 hz precision for QB compatibility */
    const double precision = 65536.0 / 0x1234DD;
    //const double precision = 0.5;
    
    return floor(result / precision) * precision;
}

BAS_STR B_FuncDATE_S()
{
    // mm-dd-yyyy
    time_t t = time(0);
    struct tm* tm = localtime(&t);
    char Buf[512];
    strftime(Buf, 512, "%m-%d-%Y", tm);
    return Buf;
}

BAS_STR B_FuncTIME_S()
{
    // hh:mm:ss
    time_t t = time(0);
    struct tm tm; localtime_r(&t,&tm);
    char Buf[512];
    strftime(Buf, 512, "%H:%M:%S", &tm);
    return Buf;
}

static void TrySetTime(const std::string& str)
{
    int month=0,
          day=0,
         year=0,
         hour=0,
       minute=0,
       second=0;
    const char* s = str.c_str();
    int n = sscanf(s,
        "%d%*[-/]%d%*[-/]%d %d:%d:%d",
        &month,
        &day,
        &year,
        &hour,
        &minute,
        &second);
    if(year >= 0 && year < 100) year += 1900;
    
    if(n < 3) { Error: B_ERROR(5); // Illegal function call 
                return; }
    
    struct tm tm;
    tm.tm_sec  = second;
    tm.tm_min  = minute;
    tm.tm_hour = hour;
    tm.tm_mday = day;
    tm.tm_mon  = month - 1;
    tm.tm_year = year  - 1900;
    time_t t = mktime(&tm);
    if(t == (time_t) -1) goto Error;
    int e = stime(&t);
    if(e == -1) B_ERROR_Errno();
}

void B_StmtTIME_S(const BAS_STR& t)
{
    // accepted formats: hh, hh:mm, hh:mm:ss
    TrySetTime(B_FuncDATE_S() + " " + t);
}

void B_StmtDATE_S(const BAS_STR& d)
{
    // accepted formats:  mm-dd-yy, mm-dd-yyyy, mm/dd/yy, mm/dd/yyyy
    TrySetTime(d + " " + B_FuncTIME_S());
}
