#include <iostream>
#include <sstream>

#include "typenameof.hh"
#include "stmt.hh"

stmt::stmt(const LexLocation& ll): lexloc(ll)
{
}
stmt::~stmt()
{
}

void stmt::Compile()
{
    std::cout << "    /* Compilation of "
              << typenameof(*this)
              << " not supported */\n";
}

static bool IsEntirelyIntegerCase(exprp exp)
{
    if(!exp) return false;

    if(dynamic_cast<expr_caseelse*>(&*exp))
    {
        return true;
    }
    if(expr_caseor* e = dynamic_cast<expr_caseor*>(&*exp))
    {
        return IsEntirelyIntegerCase(e->a) && IsEntirelyIntegerCase(e->b);
    }
    if(expr_caseis* e = dynamic_cast<expr_caseis*>(&*exp))
    {
        if(e->token != (QBtoken)'=') return false;
        return IsEntirelyIntegerCase(e->ref);
    }
    if(expr_caserange* e = dynamic_cast<expr_caserange*>(&*exp))
    {
        return IsEntirelyIntegerCase(e->min) && IsEntirelyIntegerCase(e->max);
    }

    if(!exp->is_const()) return false;

    if(expr_number* e = dynamic_cast<expr_number*>(&*exp))
        if(e->is_int()) return true;

    TypeP type = exp->GetType();
    return type && type->is_integer();
}

bool stmt_select::IsEntirelyGotoCase() const
{
    for(const auto& c: cases)
    {
        const stmtp& s = c.second;
        const stmt_labelref* l = dynamic_cast<const stmt_labelref*> (&*s);
        if(!l || l->token != GOTO) return false;
    }
    return true;
}

bool stmt_select::IsEntirelyGosubCase() const
{
    for(const auto& c: cases)
    {
        const stmtp& s = c.second;
        const stmt_labelref* l = dynamic_cast<const stmt_labelref*> (&*s);
        if(!l || l->token != GOSUB) return false;
    }
    return true;
}

void stmt_print::OptimizeTail()
{
    std::string buffered_string;
    while(!params.empty())
    {
        std::pair<QBtoken,exprp>& last = params.back();
        if(last.first != (QBtoken)0) break;

        if(expr_string* str = dynamic_cast<expr_string*>(&*last.second))
        {
            buffered_string.insert(0, str->get());
            params.erase(--params.end());
        }
        else if(expr_number* num = dynamic_cast<expr_number*>(&*last.second))
        {
            if(!num->is_int()) break;
            std::stringstream tmp; tmp << num->get_i();
            buffered_string.insert(0, tmp.str());
            params.erase(--params.end());
        }
        else break;
    }
    if(!buffered_string.empty())
        params.push_back(std::make_pair((QBtoken)0, new expr_string(buffered_string)));
}

bool stmt_select::IsEntirelyIntegerCase() const
{
    for(const auto& c: cases)
    {
        if(!::IsEntirelyIntegerCase(c.first)) return false;
    }
    return true;
}

void stmt_if::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    func(thenp);
    for(const auto& e: elseifp)
        func(e.second);
    func(elsep);
    stmt::ForEachSubStmt(func);
}
void stmt_if::ForEachSubExpr(const expr_visitor& func)
{
    func(condition);
    for(const auto& e: elseifp)
        func(e.first);
    stmt::ForEachSubExpr(func);
}

void stmt_select::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    for(const auto& c: cases)
        func(c.second);
    stmt::ForEachSubStmt(func);
}
void stmt_select::ForEachSubExpr(const expr_visitor& func)
{
    func(ref);
    for(const auto& c: cases)
        func(c.first);
    stmt::ForEachSubExpr(func);
}

void stmt_doloop::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    func(bodyp);
    stmt::ForEachSubStmt(func);
}
void stmt_doloop::ForEachSubExpr(const expr_visitor& func)
{
    func(precondition);
    func(postcondition);
    stmt::ForEachSubExpr(func);
}

void stmt_for::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    func(bodyp);
    stmt::ForEachSubStmt(func);
}
void stmt_for::ForEachSubExpr(const expr_visitor& func)
{
    func(var, true);
    func(begin);
    func(end);
    func(step);
    stmt::ForEachSubExpr(func);
}


void stmt_labelref::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_labelref::ForEachSubExpr(const expr_visitor& func)
{
    func(label);
    stmt::ForEachSubExpr(func);
}

void stmt_read::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_read::ForEachSubExpr(const expr_visitor& func)
{
    func(target, true);
    stmt::ForEachSubExpr(func);
}


void stmt_let::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_let::ForEachSubExpr(const expr_visitor& func)
{
    func(target, true);
    func(value);
    stmt::ForEachSubExpr(func);
}

void stmt_print::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_print::ForEachSubExpr(const expr_visitor& func)
{
    func(target);
    func(using_string);
    for(const auto& p: params)
        func(p.second);
    stmt::ForEachSubExpr(func);
}

void stmt_input::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_input::ForEachSubExpr(const expr_visitor& func)
{
    func(source);
    func(prompt);
    for(const auto& p: params)
        func(p, true);
    stmt::ForEachSubExpr(func);
}


void stmt_fileopen::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_fileopen::ForEachSubExpr(const expr_visitor& func)
{
    func(filename);
    func(mode_str);
    func(reclen);
    stmt::ForEachSubExpr(func);
}

void stmt_redim::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_redim::ForEachSubExpr(const expr_visitor& func)
{
    func(target, true);
    for(const auto& d: dimensions)
    {
        func(d.first);
        func(d.second);
    }
    stmt::ForEachSubExpr(func);
}


void stmt_erase::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_erase::ForEachSubExpr(const expr_visitor& func)
{
    func(target, true);
    stmt::ForEachSubExpr(func);
}

void stmt_subcall::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_subcall::ForEachSubExpr(const expr_visitor& func)
{
    for(std::size_t a=0; a<params.size(); ++a)
    {
        func(params[a], maybe, name, a);
    }
    stmt::ForEachSubExpr(func);
}


void stmt_label::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_label::ForEachSubExpr(const expr_visitor& func)
{
    stmt::ForEachSubExpr(func);
}


void stmt_lib_stmt::ForEachSubStmt(const std::function<void(stmtp)>& func)
{
    stmt::ForEachSubStmt(func);
}
void stmt_lib_stmt::ForEachSubExpr(const expr_visitor& func)
{
    for(std::size_t a=0; a<params.size(); ++a)
        func(params[a], lvaluemap.size() > a && lvaluemap[a]);
    stmt::ForEachSubExpr(func);
}
